!function($){

  // DateTimeWidget
  // --------------

  var DateTimeWidget = function(element, options){
    this.init(element, options)
  }

  DateTimeWidget.prototype = {

    adders: {
      "d": "addDays"
      , "M": "addMonths"
      , "y": "addYears"
      , "H": "addHours"
      , "m": "addMinutes"
    }

    , init: function(element, options){

      this.element = element
      this.$element = $(element)
      this.options = options

      var parsed = this.parseFormat()
      this.segments = parsed.segments
      this.order = parsed.order

      var date = this.$element.val() && this.parseDate(this.$element.val())
      this.date = date || this.options.date || new Date()

      this.current = undefined

      this.$element.on('click', bind(this.onClick, this))
      this.$element.on('keydown', bind(this.onKeypress, this))

      this.render()
    }

    , render: function(){
        this.$element.val(this.formatted())
        if(this.current) selection(this.element, this.current.pos[0], this.current.pos[1])
    }

    , parseFormat: function(){
        var segments = ['y', 'M', 'd', 'H', 'm']
          , date = {}
          , iformat = 0
          , format = this.options.format
          , current, last, order = []
       while(iformat < format.length){
         var c = format.charAt(iformat)
         var token = ""
         while(format.charAt(iformat) == c && iformat < format.length){
           token += format.charAt(iformat++)
         }
         if(segments.indexOf(c) >= 0){
           date[c] = {
             pos: [iformat-token.length, iformat]
             , char: c
           }
           current = date[c]
           if(last){
             current._prev = last
             last._next = current
           }
           last = current
           order.push(c)
         }
       }

      return {
        segments: date
        , order: order
      }
    }

    , parseDate: function(value){
      var date = {}
      $.each(this.segments, function(key, seg){
        date[key] = parseInt(value.slice(seg.pos[0], seg.pos[1])) || 0
      })
      return new Date(date.y, date.M-1, date.d, date.H, date.m)
    }

    , formatted: function(format){
      var format = format || this.options.format
      format = format.replace('yyyy', this.date.getFullYear());
      format = format.replace('yy', String(this.date.getFullYear()).slice(-2))
      format = format.replace('MM', this.pad2(this.date.getMonth() + 1))
      format = format.replace('dd', this.pad2(this.date.getDate()))
      format = format.replace('HH', this.pad2(this.date.getHours()))
      format = format.replace('mm', this.pad2(this.date.getMinutes()))
      return format
    }

    , first: function(){
      return this.segments[this.order[0]]
    }

    , last: function(){
      return this.segments[this.order.slice(-1)]
    }

    , next: function(){
      var next = this.current._next
      if(next) this.current = next
      else this.current = this.first()
      return this.current
    }

    , prev: function(){
      var prev = this.current._prev
      if(prev) this.current = prev
      else this.current = this.last()
      return this.current
    }

    , addDays: function(delta){
      var day = this.date.getDate() + delta
      this.date.setDate(day)
      this.render()
    }

    , addMonths: function(delta){
      var month = this.date.getMonth() + delta
      this.date.setMonth(month)
      this.render()
    }

    , addYears: function(delta){
      var year = this.date.getFullYear() + delta
      this.date.setFullYear(year)
      this.render()
    }

    , addHours: function(delta){
      var hours = this.date.getHours() + delta
      this.date.setHours(hours)
      this.render()
    }

    , addMinutes: function(delta){
      var minutes = this.date.getMinutes() + delta
      this.date.setMinutes(minutes)
      this.render()
    }

    // UI

    , select: function(seg){
      if(seg){
        selection(this.element, seg.pos[0], seg.pos[1])
      }
    }

    // handlers

    , handleSelection: function(){
        var range = selection(this.element)
        $.each(this.segments, bind(function(key, seg){
          if(between(range.start, seg.pos[0], seg.pos[1])){
            this.current = seg
            this.select(seg)
            return false
          }
        }, this))
    }

    , onClick: function(event){
        this.handleSelection()
    }

    // XXX: refactor to use keypress and keydown
    , onKeypress: function(event){
        event.preventDefault()
        // Arrow Down
        if(event.which === 40){
          this[this.adders[this.current.char]](-1)
        }
        // Arrow Up
        else if(event.which === 38){
          this[this.adders[this.current.char]](1)
        }
        // Arrow Right
        else if(event.which === 39){
          this.select(this.next())
        }
        // Arrow Left
        else if(event.which === 37){
          this.select(this.prev())
        }
        // Skip chars
        // <right> - . , <space>
        else if([39, 189, 190, 188, 32].indexOf(event.which) >= 0){
          this.select(this.next())
        }
        // <esc>
        else if(event.which === 27){
          this.$element.blur()
        }
        // digit: 0 - 9
        else if (between(event.which, 48, 57)){
          var digit = event.which - 48
          console.log(digit)
        }
    }

    , pad2: function(number){
        return ('0' + number).slice(-2)
    }

  }

  // utils

  var between = function(number, first, last){
    // we dont need to handle negative values
    return number >= first && number <= last
  }

  var selection = function(el, start, end){
    if(start >= 0){
      el.selectionStart = start
      el.selectionEnd = end || start
      el.focus()
    }
    return {start: el.selectionStart, end: el.selectionEnd}
  }

  var isFunction = function(obj){
    return typeof obj === 'function'
  }
  var nativeBind = Function.prototype
  var slice = Array.prototype.slice
  var bind = function bind(func, context) {
    var bound, args;
    if (func.bind === nativeBind && nativeBind) return nativeBind.apply(func, slice.call(arguments, 1));
    if (!isFunction(func)) throw new TypeError;
    args = slice.call(arguments, 2);
    return bound = function() {
      if (!(this instanceof bound)) return func.apply(context, args.concat(slice.call(arguments)));
      ctor.prototype = func.prototype;
      var self = new ctor;
      var result = func.apply(self, args.concat(slice.call(arguments)));
      if (Object(result) === result) return result;
      return self;
    };
  };

  $.fn.dtwidget = function(args){
    return this.each(function () {
      var $this = $(this)
        , data = $this.data('dtwidget')
        , options = $.extend({}
                             , {format: "dd.MM.yyyy HH:mm"}
                             , typeof args == 'object' && args)
      if (!data) $this.data('dtwidget', (data = new DateTimeWidget(this, options)))
      if (typeof option == 'string') data[option]()
    })
  }

  $.fn.dtwidget.Constructor = DateTimeWidget

}(window.jQuery);
